import datetime
from dateutil import tz
import dateutil.parser as p
import time
import ciso8601

#TODO: попробовать модуль isotoint

# функция преобразовывает пришедшую метку времени в количество микросекунд
microsec = 1000000
start_ts = datetime.datetime.fromtimestamp(0, datetime.timezone.utc)

def ts (time_data: int | str | None = None) -> int:
    """Функция возвращает метку времени как целое число микросекунд,
    прошедших с 1 января 1970 г. UTC

    :param time_data: Входящая метка времени. По умолчанию - None.
        Если целое число - будет возвращена она же,
        если строка - будет преобразована к целому числу микросекунд.
        Если строка не содержит даты, то вместо даты будут выбраны текущие
        сутки.
    :type time_data: int | str = None
    :returns: Целое число микросекунд с 01.01.1970
    :rtype: int
    """
    if time_data is None:
        return now_int()
    elif isinstance (time_data, int):
        return time_data

    def parse_ts(string: str) -> datetime.datetime:
        timestampFrom = ciso8601.parse_datetime (string)
        if timestampFrom.tzinfo is None:
            timestampFrom = timestampFrom.replace (tzinfo=datetime.timezone.utc)

        return timestampFrom

    try:
        timestampFrom = parse_ts(time_data)
    except ValueError as _:
        # предполагаем, что строка не содержит даты, только время
        timestampFrom = parse_ts(str(p.parse(time_data, default=datetime.datetime.now())))

    return int ((timestampFrom - start_ts).total_seconds() * microsec)

def int_to_local_timestamp (int_ts: int) -> datetime.datetime:
    if int_ts is None:
        return None

    return datetime.datetime.fromtimestamp(int_ts / microsec, tz.tzlocal())

def ts_to_local_str (ts: int | str) -> str:
    if ts is None:
        return None
    if isinstance(ts, str):
        return ts

    return str(datetime.datetime.fromtimestamp(ts / microsec, tz.tzlocal()))

def now_int() -> int:
    """ Количество микросекунд, начиная с 1 января 1970 UTC
    :rtype: int
    """
    return int(time.time() * microsec)



"""
Базовый модуль для создания коннекторов Peresvet

Экспортирует основные классы и исключения:
- BaseConnector - базовый класс коннектора
- Конфигурационные модели
- Кастомные исключения
"""

from .connector import (
    BaseConnector,
    TagGroupReaderConnector,
    CN_Q_GOOD,
    CN_Q_UNLINK_CONNECTOR_TO_SOURCE,
    CN_Q_SOURCE_ERROR,
    main
)

from .config import (
    ConnectorConfig,
    PlatformConfig,
    LogConfig,
    SSLConfig,
    ConnectorPrsJsonConfigStringFromPlatform
)

from .exceptions import (
    ConnectorBaseError,
    PlatformConnectionError,
    ConfigValidationError,
    DataProcessingError,
    PlatformConfigError
)

from .times import (
    ts,
    int_to_local_timestamp,
    ts_to_local_str,
    now_int
)
from importlib.metadata import version as _pkg_version, PackageNotFoundError

from typing_extensions import Self
__all__ = [
    'Self',
    'BaseConnector',
    'TagGroupReaderConnector',
    'ConnectorConfig',
    'PlatformConfig',
    'LogConfig',
    'SSLConfig',
    'ConnectorBaseError',
    'PlatformConnectionError',
    'ConnectorPrsJsonConfigStringFromPlatform',
    'ConfigValidationError',
    'DataProcessingError',
    'PlatformConfigError',
    'ts',
    'int_to_local_timestamp',
    'ts_to_local_str',
    'now_int',
    'CN_Q_GOOD',
    'CN_Q_UNLINK_CONNECTOR_TO_SOURCE',
    'CN_Q_SOURCE_ERROR',
    'main'
]

try:
    from ._version import version as __version__
except Exception:
    try:
        __version__ = _pkg_version("prs_connector_core")
    except PackageNotFoundError:
        __version__ = "0.0.0.dev0"


import json
import asyncio
from uuid import uuid4
from types import SimpleNamespace
from typing import Any, cast

import pytest
import aiomqtt

from prs_connector_core.config import TagAttributes, TagPrsJsonConfigStringFromPlatform
from prs_connector_core.connector import BaseConnector, TagGroupReaderConnector, CN_Q_UNLINK_CONNECTOR_TO_SOURCE


class TestConnector(BaseConnector):
    async def _read_tags(self):
        return None


class DummyTagGroupConnector(TagGroupReaderConnector):
    def __init__(self, config_file: str = "config.json") -> None:
        super().__init__(config_file=config_file)
        self.connect_result = True
        self.read_group_calls = []
        self.close_source_calls = 0

    async def _connect_to_source(self) -> bool:
        return self.connect_result

    async def _close_source(self):
        self.close_source_calls += 1
        self._source_connected.clear()

    async def _read_group(self, frequency: float):
        self.read_group_calls.append(frequency)


class DummyJsonata:
    def __init__(self, result):
        self._result = result

    def evaluate(self, _):
        return self._result


def _make_connector(tmp_path, monkeypatch) -> TestConnector:
    monkeypatch.chdir(tmp_path)
    config_file = tmp_path / "config.json"
    config_file.write_text(json.dumps({
        "id": "550e8400-e29b-41d4-a716-446655440000",
        "url": "mqtt://localhost",
    }))
    return TestConnector(str(config_file))


def _register_tag(
    conn: TestConnector,
    tag_id: str,
    value_type: int,
    max_dev: float = 0,
):
    conn._config_from_platfrom.tags[tag_id] = TagAttributes(
        prsActive=True,
        prsValueTypeCode=value_type,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(maxDev=max_dev),
    )
    conn._tag_cache[tag_id] = {"JSONataExpr": None, "lastValue": None}


def _make_tag_group_connector(tmp_path, monkeypatch) -> DummyTagGroupConnector:
    monkeypatch.chdir(tmp_path)
    config_file = tmp_path / "config.json"
    config_file.write_text(json.dumps({
        "id": "550e8400-e29b-41d4-a716-446655440000",
        "url": "mqtt://localhost",
    }))
    return DummyTagGroupConnector(str(config_file))


def test_connector_no_config_file():
    with pytest.raises(RuntimeError):
        TestConnector("nonexistent_config")


def test_connector_config_file(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    assert str(conn._config_from_file.id) == "550e8400-e29b-41d4-a716-446655440000"
    assert conn._config_from_file.url == "mqtt://localhost"


def test_process_tags_data_float_conversion_and_new_order(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    _register_tag(conn, tag_id, value_type=1, max_dev=0)

    payload = {"data": [{"tagId": tag_id, "data": [[123456, "12.5", 100]]}]}
    result = conn._process_tags_data(payload)

    assert result == {"data": [{"tagId": tag_id, "data": [[123456, 12.5, 100]]}]}
    assert conn._tag_cache[tag_id]["lastValue"] == [123456, 12.5, 100]


def test_process_tags_data_value_only_gets_timestamp_in_first_position(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    _register_tag(conn, tag_id, value_type=1, max_dev=0)
    monkeypatch.setattr("prs_connector_core.connector.now_int", lambda: 222222)

    payload = {"data": [{"tagId": tag_id, "data": [["7.5"]]}]}
    result = conn._process_tags_data(payload)

    assert result == {"data": [{"tagId": tag_id, "data": [[222222, 7.5]]}]}


def test_process_tags_data_uses_jsonata_on_value_index(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    _register_tag(conn, tag_id, value_type=2, max_dev=0)
    conn._tag_cache[tag_id]["JSONataExpr"] = DummyJsonata(result="converted")

    payload = {"data": [{"tagId": tag_id, "data": [[1000, "raw", 100]]}]}
    result = conn._process_tags_data(payload)

    assert result == {"data": [{"tagId": tag_id, "data": [[1000, "converted", 100]]}]}


def test_process_tags_data_respects_max_dev_for_numeric_values(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    _register_tag(conn, tag_id, value_type=1, max_dev=5)
    conn._tag_cache[tag_id]["lastValue"] = [1000, 10.0, 100]

    payload = {"data": [{"tagId": tag_id, "data": [[1001, 12.0, 100], [1002, 16.0, 100]]}]}
    result = conn._process_tags_data(payload)

    assert result == {"data": [{"tagId": tag_id, "data": [[1002, 16.0, 100]]}]}


def test_process_tags_data_emits_on_quality_change(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    _register_tag(conn, tag_id, value_type=1, max_dev=10)
    conn._tag_cache[tag_id]["lastValue"] = [1000, 10.0, 100]

    payload = {"data": [{"tagId": tag_id, "data": [[1001, 10.0, 103]]}]}
    result = conn._process_tags_data(payload)

    assert result == {"data": [{"tagId": tag_id, "data": [[1001, 10.0, 103]]}]}


def test_process_tags_data_json_compares_dict_content_not_key_order(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    _register_tag(conn, tag_id, value_type=4, max_dev=1)
    conn._tag_cache[tag_id]["lastValue"] = [1000, {"a": 1, "b": 2}, 100]

    payload = {"data": [{"tagId": tag_id, "data": [[1001, {"b": 2, "a": 1}, 100]]}]}
    result = conn._process_tags_data(payload)

    assert result == {"data": []}


@pytest.mark.asyncio
async def test_create_tag_cache_skips_inactive_tag(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    conn._config_from_platfrom.tags[tag_id] = TagAttributes(
        prsActive=False,
        prsValueTypeCode=1,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(),
    )

    created = await conn._create_tag_cache(tag_id)
    assert created is False
    assert tag_id not in conn._tag_cache


@pytest.mark.asyncio
async def test_create_tag_cache_with_jsonata(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    conn._config_from_platfrom.tags[tag_id] = TagAttributes(
        prsActive=True,
        prsValueTypeCode=1,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(JSONata="$sum([1,2])"),
    )

    created = await conn._create_tag_cache(tag_id)
    assert created is True
    assert conn._tag_cache[tag_id]["lastValue"] is None
    assert conn._tag_cache[tag_id]["JSONataExpr"] is not None


@pytest.mark.asyncio
async def test_remove_tag_cache(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    conn._tag_cache[tag_id] = {"lastValue": [1, 2, 100], "JSONataExpr": None}

    await conn._remove_tag_cache(tag_id)
    assert tag_id not in conn._tag_cache


def test_dict_helpers_are_order_independent():
    d1 = {"a": 1, "b": 2}
    d2 = {"b": 2, "a": 1}
    d3 = {"a": 1, "b": 3}

    assert BaseConnector._hash_dict(d1) == BaseConnector._hash_dict(d2)
    assert BaseConnector._dicts_are_equal(d1, d2) is True
    assert BaseConnector._dicts_are_equal(d1, d3) is False


@pytest.mark.asyncio
async def test_get_connector_configuration_updates_and_restarts_reading(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)

    setup_logger_called = {"value": 0}
    save_called = {"value": 0}

    async def long_read_tags():
        await asyncio.sleep(10)

    conn._read_tags_task = asyncio.create_task(asyncio.sleep(10))
    monkeypatch.setattr(conn, "_read_tags", long_read_tags)
    monkeypatch.setattr(conn, "_setup_logger", lambda: setup_logger_called.__setitem__("value", setup_logger_called["value"] + 1))
    monkeypatch.setattr(
        "prs_connector_core.connector.PlatformConfig.save",
        lambda self, _id: save_called.__setitem__("value", save_called["value"] + 1),
    )

    mes = {
        "data": {
            "prsActive": True,
            "prsEntityTypeCode": 7,
            "prsJsonConfigString": {
                "source": {"host": "127.0.0.1"},
                "log": {"level": "DEBUG", "maxBytes": 1234, "backupCount": 2},
            },
        }
    }

    await conn._get_connector_configuration_from_platform(mes)

    assert conn._config_from_platfrom.prsEntityTypeCode == 7
    assert conn._config_from_platfrom.prsJsonConfigString.source == {"host": "127.0.0.1"}
    assert conn._config_from_platfrom.prsActive is True
    assert setup_logger_called["value"] == 1
    assert save_called["value"] == 1
    assert conn._read_tags_task is not None

    conn._read_tags_task.cancel()
    await asyncio.gather(conn._read_tags_task, return_exceptions=True)


@pytest.mark.asyncio
async def test_get_connector_configuration_without_changes_does_not_save(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)

    setup_logger_called = {"value": 0}
    save_called = {"value": 0}
    monkeypatch.setattr(conn, "_setup_logger", lambda: setup_logger_called.__setitem__("value", setup_logger_called["value"] + 1))
    monkeypatch.setattr(
        "prs_connector_core.connector.PlatformConfig.save",
        lambda self, _id: save_called.__setitem__("value", save_called["value"] + 1),
    )

    mes = {
        "data": {
            "prsActive": conn._config_from_platfrom.prsActive,
            "prsEntityTypeCode": conn._config_from_platfrom.prsEntityTypeCode,
            "prsJsonConfigString": {
                "source": conn._config_from_platfrom.prsJsonConfigString.source,
                "log": conn._config_from_platfrom.prsJsonConfigString.log.model_dump(),
            },
        }
    }

    await conn._get_connector_configuration_from_platform(mes)
    assert setup_logger_called["value"] == 0
    assert save_called["value"] == 0


@pytest.mark.asyncio
async def test_tags_add_or_changed_full_list_adds_and_removes_tags(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    existing_tag = str(uuid4())
    removed_tag = str(uuid4())
    new_tag = str(uuid4())

    conn._config_from_platfrom.tags[existing_tag] = TagAttributes(
        prsActive=True,
        prsValueTypeCode=1,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(maxDev=0),
    )
    conn._config_from_platfrom.tags[removed_tag] = TagAttributes(
        prsActive=True,
        prsValueTypeCode=1,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(maxDev=0),
    )

    removed = []
    refreshed = {"value": 0}
    saved = {"value": 0}

    async def fake_remove_tag(tag_id):
        removed.append(tag_id)
        conn._config_from_platfrom.tags.pop(tag_id, None)

    async def fake_create_tag_cache(_tag_id):
        return True

    async def fake_refresh():
        refreshed["value"] += 1

    monkeypatch.setattr(conn, "_remove_tag", fake_remove_tag)
    monkeypatch.setattr(conn, "_create_tag_cache", fake_create_tag_cache)
    monkeypatch.setattr(conn, "_refresh_read_tags", fake_refresh)
    monkeypatch.setattr(
        "prs_connector_core.connector.PlatformConfig.save",
        lambda self, _id: saved.__setitem__("value", saved["value"] + 1),
    )

    mes = {
        "data": {
            "tags": {
                existing_tag: {
                    "prsActive": True,
                    "prsValueTypeCode": 1,
                    "prsJsonConfigString": {"maxDev": 0},
                },
                new_tag: {
                    "prsActive": True,
                    "prsValueTypeCode": 2,
                    "prsJsonConfigString": {"maxDev": 0},
                },
            }
        }
    }

    await conn._tags_add_or_changed(mes, full_list=True)

    assert removed == [removed_tag]
    assert set(conn._config_from_platfrom.tags.keys()) == {existing_tag, new_tag}
    assert refreshed["value"] == 1
    assert saved["value"] == 1


@pytest.mark.asyncio
async def test_tags_deleted_removes_tags_and_saves(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_1 = str(uuid4())
    tag_2 = str(uuid4())
    conn._config_from_platfrom.tags[tag_1] = TagAttributes(
        prsActive=True,
        prsValueTypeCode=1,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(maxDev=0),
    )
    conn._config_from_platfrom.tags[tag_2] = TagAttributes(
        prsActive=True,
        prsValueTypeCode=1,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(maxDev=0),
    )

    removed_cache = []
    saved = {"value": 0}

    async def fake_remove_tag_cache(tag_id):
        removed_cache.append(tag_id)

    monkeypatch.setattr(conn, "_remove_tag_cache", fake_remove_tag_cache)
    monkeypatch.setattr(
        "prs_connector_core.connector.PlatformConfig.save",
        lambda self, _id: saved.__setitem__("value", saved["value"] + 1),
    )

    await conn._tags_deleted({"data": {"tags": [tag_1]}})

    assert removed_cache == [tag_1]
    assert tag_1 not in conn._config_from_platfrom.tags
    assert tag_2 in conn._config_from_platfrom.tags
    assert saved["value"] == 1


@pytest.mark.asyncio
async def test_command_executes_all_lines(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    executed = []
    monkeypatch.setattr("prs_connector_core.connector.os.system", lambda cmd: executed.append(cmd) or 0)

    await conn._command({"data": {"command": {"lines": ["echo 1", "echo 2"]}}})
    assert executed == ["echo 1", "echo 2"]


@pytest.mark.asyncio
async def test_deleted_deactivates_connector_saves_and_shutdown(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)

    saved = {"value": 0}
    shutdown_called = {"value": 0}

    async def fake_shutdown():
        shutdown_called["value"] += 1

    monkeypatch.setattr(
        "prs_connector_core.connector.PlatformConfig.save",
        lambda self, _id: saved.__setitem__("value", saved["value"] + 1),
    )
    monkeypatch.setattr(conn, "_shutdown", fake_shutdown)

    await conn._deleted({"data": {}})

    assert conn._config_from_platfrom.prsActive is False
    assert saved["value"] == 1
    assert shutdown_called["value"] == 1


class DummyMqttClient:
    def __init__(self):
        self.calls = []

    async def publish(self, **kwargs):
        self.calls.append(kwargs)


@pytest.mark.asyncio
async def test_push_data_publishes_when_connected_and_data_present(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    dummy_client = DummyMqttClient()
    cast(Any, conn)._mqtt_client = dummy_client
    conn._mqtt_connected.set()
    conn._canceled = False

    conn._process_tags_data = lambda data: {"data": [{"tagId": "t", "data": [[1, 2, 100]]}]}
    await conn._data_queue.put({"data": [{"tagId": "t", "data": [[1, 2, 100]]}]})

    task = asyncio.create_task(conn._push_data())
    await asyncio.sleep(0.05)
    task.cancel()
    await asyncio.gather(task, return_exceptions=True)

    assert len(dummy_client.calls) == 1
    assert dummy_client.calls[0]["topic"] == "prsTag/app_api_client/data_set/*"
    assert dummy_client.calls[0]["retain"] is True


@pytest.mark.asyncio
async def test_push_data_writes_to_buffer_when_disconnected(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    conn._mqtt_connected.clear()
    conn._canceled = False
    conn._buf_file_name = str(tmp_path / "backup_test.dat")

    conn._process_tags_data = lambda data: {"data": [{"tagId": "t", "data": [[1, 2, 100]]}]}
    await conn._data_queue.put({"data": [{"tagId": "t", "data": [[1, 2, 100]]}]})

    task = asyncio.create_task(conn._push_data())
    await asyncio.sleep(0.05)
    task.cancel()
    await asyncio.gather(task, return_exceptions=True)

    written = (tmp_path / "backup_test.dat").read_text()
    assert '"tagId": "t"' in written


@pytest.mark.asyncio
async def test_process_buffer_moves_lines_to_queue_with_processed_flag(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    conn._canceled = False
    conn._mqtt_connected.set()
    conn._buf_file_name = str(tmp_path / "backup_test.dat")
    conn._tmp_buf_file_name = str(tmp_path / "backup_test.tmp")

    (tmp_path / "backup_test.dat").write_text('{"data":[{"tagId":"x","data":[[1,2,100]]}]}\n')

    pushed = []

    class QueueStub:
        def put_nowait(self, js):
            pushed.append(js)
            conn._canceled = True

    cast(Any, conn)._data_queue = QueueStub()

    await conn._process_buffer()

    assert len(pushed) == 1
    assert pushed[0]["processed"] is True


@pytest.mark.asyncio
async def test_refresh_read_tags_cancels_previous_task_and_starts_new(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)

    async def old_task():
        await asyncio.sleep(10)

    async def new_read_tags():
        await asyncio.sleep(10)

    conn._read_tags_task = asyncio.create_task(old_task())
    old_ref = conn._read_tags_task
    monkeypatch.setattr(conn, "_read_tags", new_read_tags)

    await conn._refresh_read_tags()

    assert old_ref.cancelled()
    assert conn._read_tags_task is not None
    assert conn._read_tags_task is not old_ref

    conn._read_tags_task.cancel()
    await asyncio.gather(conn._read_tags_task, return_exceptions=True)


@pytest.mark.asyncio
async def test_handle_messages_dispatches_actions(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    conn._canceled = False
    conn._mqtt_connected.set()

    called = []

    async def full(_): called.append("full")
    async def conn_cfg(_): called.append("connector")
    async def tags_cfg(_): called.append("tags")
    async def tags_del(_): called.append("deleted_tags")
    async def deleted(_): called.append("deleted")
    async def command(_): called.append("command")

    monkeypatch.setattr(conn, "_get_full_configuration_from_platform", full)
    monkeypatch.setattr(conn, "_get_connector_configuration_from_platform", conn_cfg)
    monkeypatch.setattr(conn, "_tags_add_or_changed", tags_cfg)
    monkeypatch.setattr(conn, "_tags_deleted", tags_del)
    monkeypatch.setattr(conn, "_deleted", deleted)
    monkeypatch.setattr(conn, "_command", command)

    actions = [
        "prsConnector.full_configuration",
        "prsConnector.connector_configuration",
        "prsConnector.tags_configuration",
        "prsConnector.tags_deleted",
        "prsConnector.deleted",
        "prsConnector.command",
    ]

    class FakeMessages:
        def __init__(self, payloads):
            self.payloads = payloads
            self.idx = 0

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self.idx >= len(self.payloads):
                conn._canceled = True
                raise StopAsyncIteration
            payload = self.payloads[self.idx]
            self.idx += 1
            # Топик должен начинаться с prs2conn, иначе коннектор отправит сообщение в _process_message
            return SimpleNamespace(payload=payload, topic="prs2conn/" + conn._config_from_file.id)

    payloads = [json.dumps({"action": a, "data": {}}).encode("utf8") for a in actions]
    cast(Any, conn)._mqtt_client = SimpleNamespace(messages=FakeMessages(payloads))

    await conn._handle_messages()

    assert called == ["full", "connector", "tags", "deleted_tags", "deleted", "command"]


@pytest.mark.asyncio
async def test_tag_group_create_cache_adds_tag_to_frequency_group(tmp_path, monkeypatch):
    conn = _make_tag_group_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    conn._config_from_platfrom.tags[tag_id] = TagAttributes(
        prsActive=True,
        prsValueTypeCode=1,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(frequency=2.5, maxDev=0),
    )

    created = await conn._create_tag_cache(tag_id)

    assert created is True
    assert tag_id in conn._tag_cache
    assert conn._tag_groups[2.5]["tags"] == [tag_id]


@pytest.mark.asyncio
async def test_tag_group_remove_cache_removes_group_and_cancels_task(tmp_path, monkeypatch):
    conn = _make_tag_group_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    frequency = 1.0

    conn._config_from_platfrom.tags[tag_id] = TagAttributes(
        prsActive=True,
        prsValueTypeCode=1,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(frequency=frequency, maxDev=0),
    )
    conn._tag_cache[tag_id] = {"lastValue": None, "JSONataExpr": None}
    conn._tag_groups[frequency]["tags"] = [tag_id]
    conn._tag_groups[frequency]["task"] = asyncio.create_task(asyncio.sleep(10))

    await conn._remove_tag_cache(tag_id)

    assert tag_id not in conn._tag_cache
    assert frequency not in conn._tag_groups


@pytest.mark.asyncio
async def test_tag_group_remove_cache_unknown_tag_keeps_running(tmp_path, monkeypatch):
    conn = _make_tag_group_connector(tmp_path, monkeypatch)
    unknown_tag = str(uuid4())
    conn._tag_cache[unknown_tag] = {"lastValue": None, "JSONataExpr": None}

    await conn._remove_tag_cache(unknown_tag)

    assert unknown_tag not in conn._tag_cache


@pytest.mark.asyncio
async def test_periodic_task_for_group_calls_read_group(tmp_path, monkeypatch):
    conn = _make_tag_group_connector(tmp_path, monkeypatch)
    conn._canceled = False
    conn._source_connected.set()

    async def fake_read_group(frequency):
        conn.read_group_calls.append(frequency)
        conn._canceled = True

    async def fast_sleep(_):
        return None

    monkeypatch.setattr(conn, "_read_group", fake_read_group)
    monkeypatch.setattr("prs_connector_core.connector.asyncio.sleep", fast_sleep)

    await conn._periodic_task_for_group(0.5)

    assert conn.read_group_calls == [0.5]


@pytest.mark.asyncio
async def test_read_tags_connect_false_sleeps_and_retries(tmp_path, monkeypatch):
    conn = _make_tag_group_connector(tmp_path, monkeypatch)
    conn.connect_result = False
    conn._canceled = False
    sleeps = []

    async def fast_sleep(seconds):
        sleeps.append(seconds)
        conn._canceled = True

    monkeypatch.setattr("prs_connector_core.connector.asyncio.sleep", fast_sleep)

    await conn._read_tags()

    assert sleeps == [5]


@pytest.mark.asyncio
async def test_read_tags_disconnect_pushes_bad_quality_data_and_closes_source(tmp_path, monkeypatch):
    conn = _make_tag_group_connector(tmp_path, monkeypatch)
    conn.connect_result = True
    conn._canceled = False
    conn._source_connected.clear()
    conn._tag_groups[1.0]["tags"] = ["tag-1"]
    conn._tag_cache["tag-1"] = {"lastValue": None, "JSONataExpr": None}

    queue_payloads = []

    class QueueStub:
        def put_nowait(self, item):
            queue_payloads.append(item)

    cast(Any, conn)._data_queue = QueueStub()

    async def fake_close_source():
        conn.close_source_calls += 1
        conn._canceled = True

    async def fast_sleep(_):
        return None

    monkeypatch.setattr(conn, "_close_source", fake_close_source)
    monkeypatch.setattr("prs_connector_core.connector.asyncio.sleep", fast_sleep)
    monkeypatch.setattr("prs_connector_core.connector.now_int", lambda: 777)

    await conn._read_tags()

    assert conn.close_source_calls == 1
    assert len(queue_payloads) == 1
    assert queue_payloads[0] == {
        "data": [{
            "tagId": "tag-1",
            "data": [[777, None, CN_Q_UNLINK_CONNECTOR_TO_SOURCE]],
        }]
    }


@pytest.mark.asyncio
async def test_shutdown_cancels_running_tasks(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)

    async def long_task():
        await asyncio.sleep(10)

    conn._handle_messages_task = asyncio.create_task(long_task())
    conn._push_data_task = asyncio.create_task(long_task())
    conn._process_buffer_task = asyncio.create_task(long_task())
    conn._read_tags_task = asyncio.create_task(long_task())

    await conn._shutdown()

    assert conn._handle_messages_task.cancelled()
    assert conn._push_data_task.cancelled()
    assert conn._process_buffer_task.cancelled()
    assert conn._read_tags_task.cancelled()


def test_signal_handlers_set_canceled_flag(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    conn._canceled = False

    conn._handle_signal_win(None, None)
    assert conn._canceled is True

    conn._canceled = False
    conn._handle_signal_unix()
    assert conn._canceled is True


@pytest.mark.asyncio
async def test_run_success_connect_flow(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    conn._config_from_platfrom.prsActive = False

    class FakeClientCM:
        def __init__(self):
            self.subscriptions = []
            self.publishes = []

        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_type, exc, tb):
            return False

        async def subscribe(self, topic):
            self.subscriptions.append(topic)

        async def publish(self, topic, payload, retain):
            self.publishes.append((topic, payload, retain))

    fake_client = FakeClientCM()
    shutdown_called = {"value": 0}

    async def fake_shutdown():
        shutdown_called["value"] += 1

    async def fast_sleep(seconds):
        if seconds == 3:
            conn._canceled = True
        return None

    monkeypatch.setattr("prs_connector_core.connector.aiomqtt.Client", lambda **kwargs: fake_client)
    monkeypatch.setattr("prs_connector_core.connector.asyncio.sleep", fast_sleep)
    monkeypatch.setattr("prs_connector_core.connector.signal.signal", lambda *_: None)
    monkeypatch.setattr("prs_connector_core.connector.sys.platform", "win32")
    monkeypatch.setattr(conn, "_shutdown", fake_shutdown)

    await conn.run()

    assert shutdown_called["value"] == 1
    assert fake_client.subscriptions == [conn._mqtt_topic_messages_from_platform]
    assert len(fake_client.publishes) == 1
    assert fake_client.publishes[0][0] == f"conn2prs/{conn._config_from_file.id}"


@pytest.mark.asyncio
async def test_run_handles_mqtt_error_and_retries(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    conn._config_from_platfrom.prsActive = False

    class FailingClientCM:
        async def __aenter__(self):
            raise aiomqtt.MqttError("boom")

        async def __aexit__(self, exc_type, exc, tb):
            return False

    shutdown_called = {"value": 0}

    async def fake_shutdown():
        shutdown_called["value"] += 1

    async def fast_sleep(seconds):
        if seconds == 5:
            conn._canceled = True
        return None

    monkeypatch.setattr("prs_connector_core.connector.aiomqtt.Client", lambda **kwargs: FailingClientCM())
    monkeypatch.setattr("prs_connector_core.connector.asyncio.sleep", fast_sleep)
    monkeypatch.setattr("prs_connector_core.connector.signal.signal", lambda *_: None)
    monkeypatch.setattr("prs_connector_core.connector.sys.platform", "win32")
    monkeypatch.setattr(conn, "_shutdown", fake_shutdown)

    await conn.run()

    assert shutdown_called["value"] == 1


@pytest.mark.asyncio
async def test_get_full_configuration_dispatches_to_two_handlers(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    called_connector = []
    called_tags = []

    async def fake_connector_cfg(mes):
        called_connector.append(mes)

    async def fake_tags_cfg(mes, full_list=False):
        called_tags.append((mes, full_list))

    monkeypatch.setattr(conn, "_get_connector_configuration_from_platform", fake_connector_cfg)
    monkeypatch.setattr(conn, "_tags_add_or_changed", fake_tags_cfg)

    payload = {
        "data": {
            "prsActive": False,
            "prsEntityTypeCode": 5,
            "prsJsonConfigString": {"source": {"x": 1}, "log": {}},
            "tags": {"tag-1": {"prsActive": True, "prsValueTypeCode": 1, "prsJsonConfigString": {}}},
        }
    }
    await conn._get_full_configuration_from_platform(payload)

    assert called_connector[0] == {
        "data": {
            "prsActive": False,
            "prsEntityTypeCode": 5,
            "prsJsonConfigString": {"source": {"x": 1}, "log": {}},
        }
    }
    assert called_tags[0][0] == {"data": {"tags": payload["data"]["tags"]}}
    assert called_tags[0][1] is True


@pytest.mark.asyncio
async def test_get_connector_configuration_active_toggle_to_false(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    conn._config_from_platfrom.prsActive = True
    conn._read_tags_task = asyncio.create_task(asyncio.sleep(10))
    save_called = {"value": 0}
    monkeypatch.setattr(
        "prs_connector_core.connector.PlatformConfig.save",
        lambda self, _id: save_called.__setitem__("value", save_called["value"] + 1),
    )

    mes = {
        "data": {
            "prsActive": False,
            "prsEntityTypeCode": conn._config_from_platfrom.prsEntityTypeCode,
            "prsJsonConfigString": {
                "source": conn._config_from_platfrom.prsJsonConfigString.source,
                "log": conn._config_from_platfrom.prsJsonConfigString.log.model_dump(),
            },
        }
    }

    await conn._get_connector_configuration_from_platform(mes)
    assert conn._config_from_platfrom.prsActive is False
    assert save_called["value"] == 1


@pytest.mark.asyncio
async def test_tags_add_or_changed_updates_changed_existing_tag(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    conn._config_from_platfrom.tags[tag_id] = TagAttributes(
        prsActive=True,
        prsValueTypeCode=1,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(maxDev=0),
    )
    removed = []
    refreshed = {"value": 0}
    saved = {"value": 0}

    async def fake_remove_tag(tag_id):
        removed.append(tag_id)
        conn._config_from_platfrom.tags.pop(tag_id, None)

    async def fake_create(_):
        return True

    async def fake_refresh():
        refreshed["value"] += 1

    monkeypatch.setattr(conn, "_remove_tag", fake_remove_tag)
    monkeypatch.setattr(conn, "_create_tag_cache", fake_create)
    monkeypatch.setattr(conn, "_refresh_read_tags", fake_refresh)
    monkeypatch.setattr(
        "prs_connector_core.connector.PlatformConfig.save",
        lambda self, _id: saved.__setitem__("value", saved["value"] + 1),
    )

    await conn._tags_add_or_changed({
        "data": {
            "tags": {
                tag_id: {
                    "prsActive": True,
                    "prsValueTypeCode": 2,
                    "prsJsonConfigString": {"maxDev": 0},
                }
            }
        }
    })

    assert removed == [tag_id]
    assert conn._config_from_platfrom.tags[tag_id].prsValueTypeCode == 2
    assert refreshed["value"] == 1
    assert saved["value"] == 1


@pytest.mark.asyncio
async def test_tags_add_or_changed_create_cache_failure_does_not_save(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    refreshed = {"value": 0}
    saved = {"value": 0}

    async def fake_create(_):
        return False

    async def fake_refresh():
        refreshed["value"] += 1

    monkeypatch.setattr(conn, "_create_tag_cache", fake_create)
    monkeypatch.setattr(conn, "_refresh_read_tags", fake_refresh)
    monkeypatch.setattr(
        "prs_connector_core.connector.PlatformConfig.save",
        lambda self, _id: saved.__setitem__("value", saved["value"] + 1),
    )

    await conn._tags_add_or_changed({
        "data": {
            "tags": {
                tag_id: {
                    "prsActive": True,
                    "prsValueTypeCode": 1,
                    "prsJsonConfigString": {"maxDev": 0},
                }
            }
        }
    })

    assert refreshed["value"] == 0
    assert saved["value"] == 0


@pytest.mark.asyncio
async def test_handle_messages_mqtt_error_clears_connection_flag(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    conn._mqtt_connected.set()
    conn._canceled = False

    class ErrorMessages:
        def __aiter__(self):
            return self

        async def __anext__(self):
            raise aiomqtt.MqttError("oops")

    cast(Any, conn)._mqtt_client = SimpleNamespace(messages=ErrorMessages())

    task = asyncio.create_task(conn._handle_messages())
    await asyncio.sleep(0.05)
    task.cancel()
    await asyncio.gather(task, return_exceptions=True)

    assert not conn._mqtt_connected.is_set()


@pytest.mark.asyncio
async def test_handle_messages_generic_exception_branch(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    conn._mqtt_connected.set()
    conn._canceled = False

    class ErrorMessages:
        def __aiter__(self):
            return self

        async def __anext__(self):
            conn._canceled = True
            raise ValueError("boom")

    cast(Any, conn)._mqtt_client = SimpleNamespace(messages=ErrorMessages())
    await conn._handle_messages()
    assert conn._canceled is True


@pytest.mark.asyncio
async def test_create_tag_cache_returns_false_when_jsonata_init_fails(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    conn._config_from_platfrom.tags[tag_id] = TagAttributes(
        prsActive=True,
        prsValueTypeCode=1,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(JSONata="$bad("),
    )

    def bad_jsonata(_):
        raise Exception("bad expr")

    monkeypatch.setattr("prs_connector_core.connector.Jsonata", bad_jsonata)
    created = await conn._create_tag_cache(tag_id)
    assert created is False


@pytest.mark.asyncio
async def test_tag_group_create_cache_missing_tag_after_super_returns_false(tmp_path, monkeypatch):
    conn = _make_tag_group_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())

    async def fake_super_create(self, _tag_id):
        return True

    monkeypatch.setattr(BaseConnector, "_create_tag_cache", fake_super_create)
    created = await conn._create_tag_cache(tag_id)
    assert created is False


@pytest.mark.asyncio
async def test_tag_group_remove_cache_logs_when_tag_not_in_group(tmp_path, monkeypatch):
    conn = _make_tag_group_connector(tmp_path, monkeypatch)
    tag_id = str(uuid4())
    frequency = 3.0
    conn._config_from_platfrom.tags[tag_id] = TagAttributes(
        prsActive=True,
        prsValueTypeCode=1,
        prsJsonConfigString=TagPrsJsonConfigStringFromPlatform(frequency=frequency, maxDev=0),
    )
    conn._tag_groups[frequency]["tags"] = []
    conn._tag_cache[tag_id] = {"lastValue": None, "JSONataExpr": None}

    await conn._remove_tag_cache(tag_id)
    assert frequency not in conn._tag_groups
    assert tag_id not in conn._tag_cache


@pytest.mark.asyncio
async def test_periodic_task_for_group_cancelled_branch(tmp_path, monkeypatch):
    conn = _make_tag_group_connector(tmp_path, monkeypatch)
    conn._source_connected.set()
    conn._canceled = False

    async def slow_read_group(frequency):
        await asyncio.sleep(10)

    monkeypatch.setattr(conn, "_read_group", slow_read_group)
    task = asyncio.create_task(conn._periodic_task_for_group(1.0))
    await asyncio.sleep(0.05)
    task.cancel()
    result = await asyncio.gather(task, return_exceptions=True)
    assert result[0] is None


@pytest.mark.asyncio
async def test_read_tags_source_connected_sleep_then_cancelled_branch(tmp_path, monkeypatch):
    conn = _make_tag_group_connector(tmp_path, monkeypatch)
    conn.connect_result = True
    conn._canceled = False
    conn._source_connected.set()
    conn._tag_groups[1.0]["tags"] = ["tag-1"]

    async def cancel_sleep(_):
        raise asyncio.CancelledError

    monkeypatch.setattr("prs_connector_core.connector.asyncio.sleep", cancel_sleep)

    await conn._read_tags()
    assert conn.close_source_calls == 1


@pytest.mark.asyncio
async def test_read_tags_generic_exception_branch(tmp_path, monkeypatch):
    conn = _make_tag_group_connector(tmp_path, monkeypatch)
    conn._canceled = False

    async def broken_connect():
        conn._canceled = True
        raise RuntimeError("err")

    monkeypatch.setattr(conn, "_connect_to_source", broken_connect)
    await conn._read_tags()
    assert conn._canceled is True


@pytest.mark.asyncio
async def test_process_buffer_exception_branch_exits_when_canceled(tmp_path, monkeypatch):
    conn = _make_connector(tmp_path, monkeypatch)
    conn._mqtt_connected.set()
    conn._canceled = False

    async def broken_stat(_):
        conn._canceled = True
        raise RuntimeError("stat failed")

    monkeypatch.setattr("prs_connector_core.connector.aiofiles.os.stat", broken_stat)
    await conn._process_buffer()
    assert conn._canceled is True



"""
Кастомные исключения для коннекторов Peresvet
"""

class ConnectorBaseError(Exception):
    """Базовое исключение для всех ошибок коннектора"""
    def __init__(self, message: str = "Ошибка коннектора"):
        super().__init__(message)
        self.message = message

class PlatformConnectionError(ConnectorBaseError):
    """Ошибка соединения с платформой"""
    def __init__(self, target: str):
        message = f"Не удалось подключиться к {target}"
        super().__init__(message)

class SourceConnectionError(ConnectorBaseError):
    """Ошибка соединения с источником данных"""
    def __init__(self, target: str):
        message = f"Не удалось подключиться к {target}"
        super().__init__(message)

class ConfigValidationError(ConnectorBaseError):
    """Ошибка валидации конфигурации"""
    def __init__(self, field: str, details: str):
        message = f"Некорректная конфигурация в поле {field}: {details}"
        super().__init__(message)

class DataProcessingError(ConnectorBaseError):
    """Ошибка обработки данных тега"""
    def __init__(self, tag_id: str, reason: str):
        message = f"Ошибка обработки тега {tag_id}: {reason}"
        super().__init__(message)

class PlatformConfigError(ConnectorBaseError):
    """Ошибка конфигурации платформы"""
    def __init__(self, reason: str):
        message = f"Некорректная конфигурация платформы: {reason}"
        super().__init__(message)

class JsonataError(ConnectorBaseError):
    """Ошибка выражения jsonata"""
    def __init__(self, tag_id: str, reason: str):
        message = f"Ошибка обработки jsonata '{reason}' для тега {tag_id}"
        super().__init__(message)


import pytest

'''
from prs_connector_core.connector import BaseConnector

class TestConnector(BaseConnector):
    async def read_tag(self, tag_config):
        return 42

@pytest.fixture
def connector():
    return TestConnector(config_path="tests/test_config.json")
'''


from __future__ import annotations
import os
import sys
import json
import copy
import hashlib
import logging.handlers
import ssl
import asyncio
import signal
import logging
import aiofiles.os
import aiomqtt
import aiofiles
import time
from abc import ABC, abstractmethod
from pathlib import Path
from urllib.parse import urlparse
from collections import defaultdict
from typing import Final
from multiprocessing import freeze_support

from jsonata import Jsonata
from .config import (
    ConnectorConfig,
    LogConfig,
    PlatformConfig,
    ConnectorPrsJsonConfigStringFromPlatform,
    TagAttributes
)
from .exceptions import (
    ConfigValidationError
)
from .times import now_int

CN_Q_GOOD : Final[int] = 0
CN_Q_UNLINK_CONNECTOR_TO_SOURCE : Final[int] = 102
CN_Q_SOURCE_ERROR : Final[int] = 103

class BaseConnector(ABC):
    """Базовый класс коннектора платформы Peresvet"""

    def __init__(self, config_file: str = "config.json") -> None:
        # Инициализация конфигурации из файла
        # Параметры: id, url, ssl.
        try:
            self._config_from_file : ConnectorConfig = ConnectorConfig.from_file(config_file)
        except ConfigValidationError as e:
            self._emergency_shutdown(f"Ошибка конфигурации: {e}")

        self._loop = None

        # Инициализация клиента MQTT
        self._mqtt_client : aiomqtt.Client | None = None

        # Инициализация конфигурации от платформы
        self._config_from_platfrom : PlatformConfig = PlatformConfig.from_file(self._config_from_file.id)

        # кэш тегов
        # содержит JSONata выражения и последние отправленные в платформу значения
        # имеет вид:
        # {
        #    "<tag_id>": {
        #       "JSONataExpr": Jsonata(),
        #       "lastValue": [ts, val, q]
        #    }
        # }
        self._tag_cache = {}

        self._logger : logging.Logger = None # type: ignore
        self._setup_logger()

        # очередь данных для отправки в платформу
        self._data_queue: asyncio.Queue = asyncio.Queue()
        # блокировка для работы с файлом буфера
        self._buf_file_lock: asyncio.Lock = asyncio.Lock()
        # имя файла буфера
        self._buf_file_name = f"backup_{self._config_from_file.id}.dat"
        # имя временного файла буфера
        self._tmp_buf_file_name = f"backup_{self._config_from_file.id}.tmp"
        # флаг коннекта к платформе
        self._mqtt_connected = asyncio.Event()

        # Извлекаем параметры подключения
        parsed_url = urlparse(self._config_from_file.url)

        # топик, в который платформа будет посылать сообщения для коннектора
        self._mqtt_topic_messages_from_platform = f"prs2conn/{self._config_from_file.id}"
        # Last Will: при нештатном отключении брокер публикует в этот топик — платформа запишет в теги качество 100
        self._mqtt_will_topic = f"prsConnector/connection_lost/{self._config_from_file.id}"
        self._mqtt_will_payload = json.dumps({"id": self._config_from_file.id})

        self._mqtt_parsed_url = {
            "host": parsed_url.hostname,
            "port": parsed_url.port or 1883,  # Порт по умолчанию
            "user": parsed_url.username,
            "password": parsed_url.password,
            "tls": None
        }

        try:
            # SSL, если используется mqtts://
            if self._config_from_file.ssl:
                tls_params = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                # Загружаем CA сертификат для проверки сервера
                tls_params.load_verify_locations(cafile=self._config_from_file.ssl.caFile)

                # Загружаем клиентский сертификат и приватный ключ
                tls_params.load_cert_chain(
                    certfile=self._config_from_file.ssl.certFile,
                    keyfile=self._config_from_file.ssl.keyFile
                )

                # Требуем проверку сертификатов
                tls_params.verify_mode = ssl.VerifyMode(self._config_from_file.ssl.certsRequired)
                self._mqtt_parsed_url["tls"] = tls_params
        except:
            self._emergency_shutdown("Ошибка загрузки сертификатов.")

        # обработка сообщений от платформы
        self._handle_messages_task: asyncio.Task | None = None
        # чтение данных тегов
        self._read_tags_task: asyncio.Task | None = None
        # работа с данными
        self._push_data_task: asyncio.Task | None = None
        # работа с буфером
        self._process_buffer_task: asyncio.Task | None = None

        self._canceled: bool = False

        if sys.platform.lower() == "win32" or os.name.lower() == "nt":
            # этот код работает только в Windows
            from asyncio import set_event_loop_policy, WindowsSelectorEventLoopPolicy # type: ignore
            set_event_loop_policy(WindowsSelectorEventLoopPolicy())

    async def _shutdown(self):
        """Обработчик завершения работы"""
        self._logger.info(f"Получен сигнал завершения работы, сохраняем данные...")

        # Отменяем все задачи
        tasks = [
            self._handle_messages_task,
            self._push_data_task,
            self._process_buffer_task,
            self._read_tags_task
        ]

        for task in tasks:
            if task and not task.done():
                task.cancel()

        # Ожидаем завершения задач
        _, pending = await asyncio.wait(tasks, timeout=10.0) # type: ignore

        if pending:
            self._logger.error(f"При остановке коннектора некоторые задачи не успели корректно завершиться: {pending}.")

        self._logger.info("Работа коннектора завершена.")

    def _emergency_shutdown(self, message: str) -> None:
        """Аварийное завершение работы при критических ошибках"""
        logger = logging.getLogger("prs_emergency")
        logger.error(message)
        raise RuntimeError(message)

    async def _push_data(self):
        # берём из очереди сообщение и пытаемся отправить его в платформу
        # при неудаче сохраняем в буфер

        async def write_to_buf(mes):
            self._logger.debug(f"Сообщение для записи в буфер: {mes}")
            # запись сообщения в буфер
            async with self._buf_file_lock:
                async with aiofiles.open(self._buf_file_name, mode="+a") as buf_file:
                    s = json.dumps(mes)
                    await buf_file.write(f"{s}\n")

        while not self._canceled:
            mes = await self._data_queue.get()
            new_mes = mes
            try:
                self._logger.debug(f"Новое сообщение с данными: {mes}.")
                # при помещении сообщения из буфера в очередь мы помечаем его как уже обработанное
                processed = mes.get("processed", False)
                if not processed:
                    new_mes = self._process_tags_data(mes)

                if new_mes["data"]:
                    if self._mqtt_connected.is_set():
                        self._logger.info(f"Отправка данных в платформу.")
                        await self._mqtt_client.publish( # type: ignore
                            topic="prsTag/app_api_client/data_set/*",
                            payload=json.dumps(new_mes),
                            retain=True
                        )
                    else:
                        await write_to_buf(new_mes)
                        self._logger.info(f"Данные сохранены в буфер.")
            except (aiomqtt.MqttError) as _:
                self._mqtt_connected.clear()
                if new_mes["data"]:
                    await write_to_buf(new_mes)
                    self._logger.info(f"Ошибка передачи данных: сохраняем в буфер.")
            except asyncio.CancelledError:
                return
            except Exception as ex:
                self._logger.error(f"Системная ошибка в цикле отправки данных в платформу: {ex}.")

    async def _process_buffer(self):
        # бесконечная функция обработки буфера
        # логика: если нет связи с брокером, просто ничего не делаем
        # иначе, если размер буфера > 0, то пытаемся отправить данные из него в платформу

        while not self._canceled:
            try:
                if not self._mqtt_connected.is_set():
                    await asyncio.sleep(2)
                    continue

                async with self._buf_file_lock:
                        stat = await aiofiles.os.stat(self._buf_file_name)
                        if stat.st_size > 0:
                            # если размер буфера > 0
                            self._logger.info("Обработка буфера данных.")
                            async with aiofiles.open(self._tmp_buf_file_name, mode="+a") as tmp_file:
                                queue_full = False
                                async with aiofiles.open(self._buf_file_name) as buf_file:
                                    async for line in buf_file: # type: ignore
                                        if queue_full or not self._mqtt_connected.is_set():
                                            # если в процессе обработки буфера переполнилась очередь или прервалась связь с платформой,
                                            # то все оставшиеся в буфере строки пишем во временный файл и потом
                                            # переименовываем временный файл в файл буфера
                                            self._logger.debug("Запись данных обратно в буфер.")
                                            if not line:
                                                self._logger.debug("Пустые данные.")
                                            await tmp_file.write(f"{line}\n")

                                        else:
                                            try:
                                                js = json.loads(line)
                                                js["processed"] = True
                                                self._data_queue.put_nowait(js)
                                                self._logger.debug("Запись данных из буфера в очередь.")
                                            except asyncio.QueueFull as _:
                                                if not queue_full:
                                                    self._logger.error("Очередь сообщений переполнена.")
                                                    queue_full = True
                                                    await tmp_file.write(line)
                            await aiofiles.os.replace(self._tmp_buf_file_name, self._buf_file_name)

            except asyncio.CancelledError:
                if await aiofiles.os.path.exists(self._tmp_buf_file_name):
                    await aiofiles.os.replace(self._tmp_buf_file_name, self._buf_file_name)
            except Exception as ex:
                self._logger.exception(f"Системная ошибка в цикле обработки буфера: {ex}.")
                #time.sleep(2)

    def _process_tags_data(self, data: dict) -> dict:
        """Метод обрабатывает "сырые" данные.
        Логика работы.
        На вход метода приходит массив данных, в том виде, как описано в API.
        Если у значения нет метки времени, метод её добавит
        Каждое значение тега:
        1. Конвертируется с помощью JSONata
        2. Преобразуется к нужному типу
        3. Если разница между последним посланным в платформу значением и текущим больше указанного предела,
           то значение будет помещено в очередь данных.
        Для тегов с типами данных 2(str) и 4(json) сравнение происходит следующим образом:
        если maxDev = 0, то в очередь помещается каждое новое значение тега, если maxDev > 0, то в очередь
        новое значение помещается, только если отличается от последнего записанного.

        Args:
            data(dict) - словарь с массивом данных тегов:
                {
                    "data": [
                        {
                            "tagId": "...",
                            "data": []
                        }
                    ]
                }

        Returns:
            dict - обработанные данные
        """
        new_data = {
            "data": []
        }

        try:
            cur_time = now_int()
            for tag in data["data"]:
                tag_id = tag["tagId"]
                new_tag_data = {
                    "tagId": tag_id,
                    "data": []
                }

                jsonata_expr = self._tag_cache[tag_id]["JSONataExpr"]
                last_value = self._tag_cache[tag_id]["lastValue"]
                #last_quality = self._tag_cache[tag_id]["last_q"]
                value_type = self._config_from_platfrom.tags[tag_id].prsValueTypeCode
                max_dev = self._config_from_platfrom.tags[tag_id].prsJsonConfigString.maxDev
                for data_value in tag["data"]:
                    new_data_value = list(data_value)
                    if len(new_data_value) == 1:
                        # В формате [x, y, q] при отсутствии метки времени source может вернуть только значение [y].
                        # Добавляем текущий ts в начало, чтобы привести к [x, y].
                        new_data_value.insert(0, cur_time)
                    new_data_quality = None if len(new_data_value) < 3 else new_data_value[2]
                    if jsonata_expr:
                        new_data_value[1] = jsonata_expr.evaluate(new_data_value[1])

                    if new_data_value[1] is not None:
                        match value_type:
                            case 0: new_data_value[1] = int(new_data_value[1])
                            case 1: new_data_value[1] = float(new_data_value[1])
                            case 2: new_data_value[1] = str(new_data_value[1])
                            case 4:
                                if isinstance(new_data_value[1], str):
                                    try:
                                        new_data_value[1] = json.loads(new_data_value[1])
                                    except Exception as ex:
                                        self._logger.error(f"Тег '{tag_id}'. Ошибка конвертации значения '{new_data_value[1]}' к типу {value_type}: {ex}")
                                        continue
                            case _ as code:
                                self._logger.error(f"Тег '{tag_id}'. Ошибка конвертации значения '{new_data_value[1]}' к типу {value_type}")
                                continue

                    if ((last_value is None) or # первое значение после запуска коннектора
                        (last_value[2] != new_data_quality) or # изменение качества
                        new_data_quality in (None, 100) and
                        (max_dev == 0 or
                        new_data_value[1] != last_value[1] and
                        (new_data_value[1] is None and last_value[1] is not None or
                        new_data_value[1] is not None and last_value[1] is None or
                        value_type in [0, 1] and (max_dev <= abs(last_value[1] - new_data_value[1])) or
                        value_type == 2 and last_value[1] != new_data_value[1] or
                        value_type == 4 and (
                            not isinstance(last_value[1], dict) or
                            not isinstance(new_data_value[1], dict) or
                            not self._dicts_are_equal(last_value[1], new_data_value[1])
                        )))):

                        new_tag_data["data"].append(new_data_value)
                        last_value = new_data_value
                        #last_quality = new_data_quality

                if new_tag_data["data"]:
                    self._tag_cache[tag_id]["lastValue"] = last_value
                    #self._tag_cache[tag_id]["last_q"] = last_quality
                    new_data["data"].append(new_tag_data)

        except Exception as ex:
            self._logger.error(f"Ошибка обработки данных: {ex}")

        return new_data

    def _handle_signal_win(self, signum, frame):
        """Обработчик для Windows"""
        self._canceled = True

    def _handle_signal_unix(self):
        """Обработчик для Unix"""
        self._canceled = True

    async def _subscribe_client(self) -> None :
        return

    async def run(self) -> None:

        self._loop = asyncio.get_running_loop()

        if sys.platform.lower() == "win32" or os.name.lower() == "nt":
            for sig in [signal.SIGINT, signal.SIGTERM]:
                signal.signal(sig, self._handle_signal_win)
        else:
            for sig in [signal.SIGINT, signal.SIGTERM]:
                self._loop.add_signal_handler(sig, self._handle_signal_unix)

        # создадим файл буфера
        async with aiofiles.open(self._buf_file_name, mode="+a") as _:
            pass

        for tag_id in self._config_from_platfrom.tags.keys():
            await self._create_tag_cache(tag_id)

        # обработка сообщений от платформы
        if not self._handle_messages_task:
            self._handle_messages_task = asyncio.create_task(self._handle_messages())
        # чтение данных тегов
        if self._config_from_platfrom.prsActive:
            if not self._read_tags_task:
                self._read_tags_task = asyncio.create_task(self._read_tags())
        # работа с данными
        if not self._push_data_task:
            self._push_data_task = asyncio.create_task(self._push_data())
        # работа с буфером
        if not self._process_buffer_task:
            self._process_buffer_task = asyncio.create_task(self._process_buffer())

        try:
            while not self._canceled:
                try:
                    will = aiomqtt.Will(
                        topic=self._mqtt_will_topic,
                        payload=self._mqtt_will_payload,
                        qos=0,
                        retain=False,
                    )
                    async with aiomqtt.Client(
                            identifier=self._config_from_file.id,
                            protocol=aiomqtt.ProtocolVersion.V5,
                            hostname=self._mqtt_parsed_url["host"],
                            port=self._mqtt_parsed_url["port"],
                            username=self._mqtt_parsed_url["user"],
                            password=self._mqtt_parsed_url["password"],
                            tls_params=self._mqtt_parsed_url["tls"],
                            timeout=180,
                            keepalive=180,
                            will=will,
                        ) as client:
                        self._mqtt_client = client

                        self._logger.info(f"Связь с платформой установлена.")

                        await client.subscribe(self._mqtt_topic_messages_from_platform)
                        # даём возможность наследникам подписаться на что-то ещё.
                        # необходимо, к примеру, для mqtt-коннектора
                        await self._subscribe_client()
                        self._mqtt_connected.set()
                        await asyncio.sleep(5)
                        payload = {
                            "action": "getConfig",
                            "data": {
                                "id": self._config_from_file.id
                            }
                        }
                        await client.publish(
                            f"conn2prs/{self._config_from_file.id}",
                            payload=json.dumps(payload),
                            retain=True
                        )

                        while self._mqtt_connected.is_set() and not self._canceled:
                            await asyncio.sleep(3)

                        # При штатном отключении брокер не публикует LWT (отправляется DISCONNECT).
                        # Публикуем в топик LWT вручную, чтобы платформа записала в теги качество 100.
                        if self._canceled and self._mqtt_client:
                            try:
                                await client.publish(
                                    self._mqtt_will_topic,
                                    payload=self._mqtt_will_payload,
                                    qos=0,
                                )
                            except Exception as ex:
                                self._logger.warning("Не удалось отправить сообщение о штатном отключении: %s", ex)

                except aiomqtt.MqttError as e:
                    self._logger.error(f"Разрыв связи с платформой: {e}.")
                    self._mqtt_connected.clear()
                    if not self._canceled:
                        try:
                            await asyncio.sleep(5)
                        except asyncio.CancelledError:
                            break
                    else:
                        break

            await self._shutdown()
        except asyncio.CancelledError:
            pass

        except Exception as ex:
            self._logger.exception(f"Неопределённое исключение: {ex}.")

    async def _get_full_configuration_from_platform(self, mes: dict):
        new_mes = {
            "data": {
                "prsActive": mes["data"]["prsActive"],
                "prsEntityTypeCode": mes["data"]["prsEntityTypeCode"],
                "prsJsonConfigString": mes["data"]["prsJsonConfigString"]
            }
        }
        await self._get_connector_configuration_from_platform(new_mes)

        new_mes = {
            "data": {
                "tags": mes["data"]["tags"]
            }
        }
        await self._tags_add_or_changed(mes=new_mes, full_list=True)

    @classmethod
    def _hash_dict(cls, js: dict) -> bytes:
        # Делаем хэш словаря. Функция нужна для сравнений словарей.
        dict_str = json.dumps(js, sort_keys=True, ensure_ascii=False)
        dict_bytes = dict_str.encode("utf-8")
        hasher = hashlib.sha256()
        hasher.update(dict_bytes)
        # Возвращаем шестнадцатеричное представление хэша
        return hasher.digest()

    @classmethod
    def _dicts_are_equal(cls, d1: dict, d2: dict) -> bool:
        d1_hash = cls._hash_dict(d1)
        d2_hash = cls._hash_dict(d2)
        return d1_hash == d2_hash

    async def _get_connector_configuration_from_platform(self, mes: dict):
        config_changed = False

        checked_prsJsonConfigString = ConnectorPrsJsonConfigStringFromPlatform(**mes["data"]["prsJsonConfigString"])
        log_config = checked_prsJsonConfigString.model_dump()["log"]
        lc = mes["data"]["prsJsonConfigString"].get("log")
        if not lc or not lc.get("fileName"):
            log_config["fileName"] = PlatformConfig.default_log_file_name(self._config_from_file.id)

        if not self._dicts_are_equal(
               self._config_from_platfrom.prsJsonConfigString.log.model_dump(),
               log_config):
            self._config_from_platfrom.prsJsonConfigString.log = LogConfig(**log_config)
            self._setup_logger()
            config_changed = True

        if not self._dicts_are_equal(
                self._config_from_platfrom.prsJsonConfigString.source,
                mes["data"]["prsJsonConfigString"]["source"]
            ):
            self._logger.debug(
                f"Изменена конфигурация источника данных. Было: {self._config_from_platfrom.prsJsonConfigString.source}; стало: {mes["data"]["prsJsonConfigString"]["source"]}."
            )
            self._config_from_platfrom.prsJsonConfigString.source = copy.deepcopy(mes["data"]["prsJsonConfigString"]["source"])
            config_changed = True
            if self._read_tags_task and not self._read_tags_task.done():
                self._read_tags_task.cancel()
                await asyncio.wait([self._read_tags_task])

            if mes["data"]["prsActive"]:
                self._read_tags_task = asyncio.create_task(self._read_tags())

        if mes["data"]["prsActive"] != self._config_from_platfrom.prsActive:
            self._logger.debug("Изменена активность коннектора.")

            self._config_from_platfrom.prsActive = mes["data"]["prsActive"]

            if mes["data"]["prsActive"]:
                if self._read_tags_task and not self._read_tags_task.done():
                    self._read_tags_task.cancel()
                    await asyncio.wait([self._read_tags_task])
                self._read_tags_task = asyncio.create_task(self._read_tags())
                self._logger.info("Коннектор активен, работа по чтению данных запущена.")
            else:
                if self._read_tags_task and not self._read_tags_task.done():
                    self._read_tags_task.cancel()
                    await asyncio.wait([self._read_tags_task])
                self._logger.info("Коннектор неактивен, работа по чтению данных остановлена.")

            config_changed = True

        if mes["data"]["prsEntityTypeCode"] != self._config_from_platfrom.prsEntityTypeCode:
            self._config_from_platfrom.prsEntityTypeCode = mes["data"]["prsEntityTypeCode"]
            # TODO: необходимо вызывать метод _entity_type_code_changed, но его пока нет.
            config_changed = True

        if config_changed:
            self._config_from_platfrom.save(self._config_from_file.id)
            self._logger.info("Конфигурация коннектора изменена.")

    async def _remove_tag(self, tag_id):
        """Удаление тега из списка"""
        await self._remove_tag_cache(tag_id)
        self._config_from_platfrom.tags.pop(tag_id, None)

    async def _tags_add_or_changed(self, mes: dict, full_list: bool = False):
        config_changed = False

        if full_list:
            tags_to_delete = list(set(self._config_from_platfrom.tags.keys()) - set(mes["data"]["tags"].keys()))
            if tags_to_delete:
                config_changed = True
            for tag_id in tags_to_delete:
                await self._remove_tag(tag_id)

        existing_tags = self._config_from_platfrom.tags.keys()

        for tag_id, tag_data in mes["data"]["tags"].items():
            add_tag = False
            if tag_id in existing_tags:
                # если тег уже есть в списке...
                old_tag_hash = self._hash_dict(self._config_from_platfrom.tags[tag_id].model_dump())

                tag_attrs = TagAttributes(**tag_data)
                new_tag_hash = self._hash_dict(tag_attrs.model_dump())
                if old_tag_hash != new_tag_hash:
                    add_tag = True
                    await self._remove_tag(tag_id=tag_id)
            else:
                add_tag = True

            if add_tag:
                self._config_from_platfrom.tags[tag_id] = TagAttributes(**tag_data)
                if await self._create_tag_cache(tag_id):
                    config_changed = True
                else:
                    self._logger.info(f"Неверная конфигурация тега {tag_id}.")

        if config_changed:
            self._config_from_platfrom.save(self._config_from_file.id)

            await self._refresh_read_tags()

            self._logger.info("Конфигурация тегов изменена.")

    async def _refresh_read_tags(self):
        if self._read_tags_task and not self._read_tags_task.done():
            self._read_tags_task.cancel()
            await asyncio.wait([self._read_tags_task], timeout=3)

        self._read_tags_task = asyncio.create_task(self._read_tags())

    async def _tags_deleted(self, mes: dict):
        # удаление тегов из списка обрабатываемых коннектором

        # аналогично методу _create_tag_cache, может быть переписан в классе-наследнике
        for tag_id in mes["data"]["tags"]:
            await self._remove_tag_cache(tag_id)
            self._config_from_platfrom.tags.pop(tag_id)

            self._logger.info(f"Тег {tag_id} удалён из списка.")

        self._config_from_platfrom.save(self._config_from_file.id)
        self._logger.info(f"Конфигурация сохранена.")

    async def _process_message(self, message) -> None :
        # обработчик нестандартных сообщений.
        # для классов-наследников
        return

    async def _handle_messages(self):
        while not self._canceled:
            try:
                await self._mqtt_connected.wait()
                if self._mqtt_client:
                    async for message in self._mqtt_client.messages:

                        topic = getattr(message, "topic", None) or ""
                        if str(topic).startswith("prs2conn"):
                            json_str = message.payload.decode('utf8') # type: ignore
                            message_data = json.loads(json_str)

                            self._logger.info(f"Сообщение от платформы: {message_data['action']}.")

                            match message_data["action"]:
                                case "prsConnector.full_configuration":
                                    await self._get_full_configuration_from_platform(message_data)
                                case "prsConnector.connector_configuration":
                                    await self._get_connector_configuration_from_platform(message_data)
                                case "prsConnector.tags_configuration":
                                    await self._tags_add_or_changed(message_data)
                                case "prsConnector.tags_deleted":
                                    await self._tags_deleted(message_data)
                                case "prsConnector.deleted":
                                    await self._deleted(message_data)
                                case "prsConnector.command":
                                    await self._command(message_data)
                        else:
                            await self._process_message(message)

            except aiomqtt.MqttError as ex:
                self._mqtt_connected.clear()
                self._logger.error(f"Ошибка MQTT в цикле обработки сообщений: {ex}.")
            except asyncio.CancelledError:
                return
            except Exception as ex:
                self._logger.error(f"Системная ошибка в цикле обработки сообщений: {ex}.")

    async def _command(self, message_data):
        for line in message_data["data"]["command"]["lines"]:
            os.system(line)

    async def _deleted(self, message_data):
        self._config_from_platfrom.prsActive = False
        self._config_from_platfrom.save(self._config_from_file.id)
        self._logger.info(f"Коннектор удалён из иерархии.")
        await self._shutdown()

    def _setup_logger(self):
        self._logger = logging.getLogger(f"prs_connector_{self._config_from_file.id}")
        self._logger.handlers.clear()
        self._logger.setLevel(self._config_from_platfrom.prsJsonConfigString.log.level)

        formatter = logging.Formatter(
            '%(asctime)s :: [%(levelname)s] :: %(name)s :: %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        )

        log_file = Path(self._config_from_platfrom.prsJsonConfigString.log.fileName)
        log_dir = log_file.parent
        log_dir.mkdir(parents=True, exist_ok=True)

        file_handler = logging.handlers.RotatingFileHandler(
            self._config_from_platfrom.prsJsonConfigString.log.fileName,
            maxBytes=self._config_from_platfrom.prsJsonConfigString.log.maxBytes,
            backupCount=self._config_from_platfrom.prsJsonConfigString.log.backupCount
        )
        file_handler.setFormatter(formatter)
        self._logger.addHandler(file_handler)

        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        self._logger.addHandler(console_handler)

    # ------------------------------------------------------------------------------------------------------------------
    # методы, которые можно переопределять в классах-наследниках

    async def _create_tag_cache(self, tag_id: str) -> bool:
        # в случае, если требуется кэш другого вида, необходимо переопределить
        # данный метод в классе-наследнике,
        # при этом из переопределённого метода необходимо вызвать данный метод

        if not self._config_from_platfrom.tags[tag_id].prsActive:
            # если тег неактивен, не создаём для него кэша
            return False

        self._tag_cache[tag_id] = {
            "lastValue": None, # признак того, что ещё ни одно значение тега не прочитано
            "JSONataExpr": None
        }
        expr = None
        try:
            expr = self._config_from_platfrom.tags[tag_id].prsJsonConfigString.JSONata
            if expr:
                self._tag_cache[tag_id]["JSONataExpr"] = Jsonata(expr)

            self._logger.info(f"Создан кэш для тега {tag_id}")
        except:
            self._logger.error(f"Тег {tag_id}. Ошибка создания JSONata выражения '{expr}'")
            return False

        return True

    async def _remove_tag_cache(self, tag_id: str):
        # если при удалении из конфигурации тега необходимо выполнить дополнительные действия, то
        # то данный метод необходимо переопределить в классе-наследнике
        # и вызвать данный метод
        self._tag_cache.pop(tag_id, None)

    @abstractmethod
    async def _read_tags(self):
        """Абстрактный метод для чтения тегов из источника"""
        raise NotImplementedError()

    #--------------------------------------------------------------------------------------------------------------------

class TagGroupReaderConnector(BaseConnector):
    """Класс коннектора, который читает данные из источника сам (не по подписке на события),
    при этом у каждого тега может быть своя частота чтения.
    Соответственно, у каждого тега в атрибуте prsJsonConfigString["source"] должен быть ключ frequency, тип - float,
    значение - частота чтения в секундах. Если указанного атрибута нет, значение по умолчанию принимается 5 сек.

    Коннектор формирует дополнительный кэш в атрибуте _frequency_groups:

    {
        <frequency>: {
            "tags": [<tag_id_1>, <tag_id_2>]
        }
    }

    Поэтому коннектор формирует группы тегов, при этом у каждой группы - своя частота чтения.

    Args:
        BaseConnector (_type_): _description_
    """

    def __init__(self, config_file: str = "config.json") -> None:
        super().__init__(config_file=config_file)

        self._tag_groups = defaultdict(self.default_tag_group)
        self._source_connected = asyncio.Event()

    def default_tag_group(self):
        return {"tags": [], "task": None}

    async def _create_tag_cache(self, tag_id: str) -> bool:
        if await super()._create_tag_cache(tag_id):
            try:
                frequency = self._config_from_platfrom.tags[tag_id].prsJsonConfigString.frequency
                self._tag_groups[frequency]["tags"].append(tag_id)

                return True
            except:
                self._logger.error(f"Создание кэша: указанного тега {tag_id} нет в списке.")
        return False

    async def _remove_tag_cache(self, tag_id: str):
        try:
            frequency = self._config_from_platfrom.tags[tag_id].prsJsonConfigString.frequency
        except:
            self._logger.error(f"Удаление кэша: указанного тега {tag_id} нет в списке.")
            await super()._remove_tag_cache(tag_id)
            return

        try:
            self._tag_groups[frequency]["tags"].remove(tag_id)
        except ValueError:
            self._logger.error(f"Тег {tag_id} не найден в соответствующей ему группе {frequency}.")

        if not len(self._tag_groups[frequency]["tags"]):
            if self._tag_groups[frequency]["task"] and not self._tag_groups[frequency]["task"].done():
                self._tag_groups[frequency]["task"].cancel()
                await asyncio.wait([self._tag_groups[frequency]["task"]], timeout=3)
            self._tag_groups.pop(frequency)
            self._logger.info(f"Удалена группа чтения {frequency}.")

        await super()._remove_tag_cache(tag_id)

    async def _periodic_task_for_group(self, frequency: float):
        try:
            while not self._canceled and self._source_connected.is_set():
                start = time.time()
                await self._read_group(frequency=frequency)
                duration = time.time() - start
                period = frequency - duration
                if period > 0:
                    await asyncio.sleep(period)

        except asyncio.CancelledError:
            return

    @abstractmethod
    async def _read_group(self, frequency: float):
        raise NotImplementedError()

    async def _read_tags(self):

        while not self._canceled:
            try:
                if await self._connect_to_source():
                    for frequency in self._tag_groups.keys():
                        self._tag_groups[frequency]["task"] = asyncio.create_task(
                            self._periodic_task_for_group(frequency=frequency)
                        )
                    self._logger.info(f"Задачи чтения тегов созданы.")

                    while True:
                        if self._source_connected.is_set():
                            await asyncio.sleep(2)
                        else:
                            ts = now_int()
                            data = {
                                "data": []
                            }
                            for tag_id in self._tag_cache.keys():
                                data["data"].append({
                                    "tagId": tag_id,
                                    "data": [[ts, None, CN_Q_UNLINK_CONNECTOR_TO_SOURCE]]
                                })
                            if data["data"]:
                                self._data_queue.put_nowait(data)
                                # даём время отработать задаче по отсылке сообщений в платформу
                                await asyncio.sleep(3)

                            self._logger.error(f"Разрыв связи с источником данных. Останавливаем задачи чтения данных, пытаемся восстановить связь с источником данных.")
                            tasks = []
                            for frequency in self._tag_groups.keys():
                                if self._tag_groups[frequency]["task"] and not self._tag_groups[frequency]["task"].done():
                                    tasks.append(self._tag_groups[frequency]["task"])
                            if tasks:
                                for task in tasks:
                                    task.cancel()
                                await asyncio.wait(tasks, timeout=3)
                            await self._close_source()
                            break
                else:
                    await asyncio.sleep(5)

            except asyncio.CancelledError:
                tasks = []
                for frequency in self._tag_groups.keys():
                    if self._tag_groups[frequency]["task"] and not self._tag_groups[frequency]["task"].done():
                        tasks.append(self._tag_groups[frequency]["task"])
                if tasks:
                    for task in tasks:
                        task.cancel()
                    _, pending = await asyncio.wait(tasks, timeout=3)
                    if pending:
                        self._logger.error(f"Не все задачи чтения тегов не завершились корректно.")
                    self._logger.info(f"Задачи чтения тегов остановлены.")
                await self._close_source()
                break
            except Exception as ex:
                self._logger.error(f"Системная ошибка в цикле чтения тегов: {ex}.")

    @abstractmethod
    async def _connect_to_source(self) -> bool:
        """Абстрактный метод соединения с источником"""
        raise NotImplementedError()

    @abstractmethod
    async def _close_source(self):
        """Абстрактный метод закрытия соединения с источником"""
        raise NotImplementedError()

def main(conn_cls):
    conf = 'config.json'
    if len(sys.argv) == 2:
        conf = sys.argv[1]
    try:
        conn = conn_cls(conf)
    except Exception as e:
        logging.getLogger("prs_emergency").exception("Ошибка инициализации коннектора: %s", e)
        sys.exit(1)

    if sys.platform.lower() == "win32" or os.name.lower() == "nt":
        freeze_support()
    asyncio.run(conn.run())


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from importlib.metadata import PackageNotFoundError, version as package_version

project = 'prs-connector-core'
copyright = '2025, ООО МПК'
author = 'ООО МПК'
try:
    release = package_version("prs_connector_core")
except PackageNotFoundError:
    release = "0.0.0.dev0"
version = release

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinxcontrib.httpdomain',
    'sphinxcontrib.httpexample',
    'sphinx.ext.todo',
    'sphinx_togglebutton'
]

templates_path = ['_templates']
exclude_patterns = []

language = 'ru'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
html_favicon = "_static/peresvet_middle.png"
html_logo = "pics/logo.png"
#html_theme_options = {
#    'logo_only': True,
#    'display_version': True,
#}
togglebutton_hint = ""
togglebutton_hint_hide = ""

html_theme_options = {
    'logo_only': True,
    'display_version': True,
}


from pydantic import BaseModel, field_validator, model_validator, Field, ValidationError
from pathlib import Path
from jsonata import Jsonata
import uuid
import json
from urllib.parse import urlparse
from typing import Any
from typing_extensions import Self
from .exceptions import ConfigValidationError

class SSLConfig(BaseModel):
    certFile: str
    keyFile: str
    caFile: str = ""
    certsRequired: int | str = 'CERTS_NONE'

    @field_validator('certsRequired', mode='before')
    @classmethod
    def validate_id(cls, v: str) -> int:
        match v:
            case 'CERTS_NONE': return 0
            case 'CERTS_OPTIONAL': return 1
            case 'CERTS_REQUIRED': return 2
            case _:
                raise ConfigValidationError(field='certsRequired', details="certRequired должен быть `CERTS_NONE`, `CERTS_OPTIONAL` или `CERTS_REQUIRED`")

class LogConfig(BaseModel):
    level: str = "INFO"
    fileName: str = "logs/prs_connector.log"
    maxBytes: int = 10 * 1024 * 1024  # 10MB
    backupCount: int = 10

class ConnectorPrsJsonConfigStringFromPlatform(BaseModel):
    source: dict = {}
    log: LogConfig = LogConfig()

class TagPrsJsonConfigStringFromPlatform(BaseModel):
    source: dict = {}
    maxDev: float = 0
    JSONata: str | None = None
    frequency: float | None = 5

class ConnectorConfig(BaseModel):
    id: str
    url: str
    ssl: SSLConfig | None = None

    @field_validator('id', mode='before')
    @classmethod
    def validate_id(cls, v: str) -> str:
        try:
            uuid.UUID(str(v))
            return v
        except ValueError as e:
            raise ConfigValidationError(field='id', details="id должен быть в виде GUID")

    @field_validator('url')
    @classmethod
    def validate_url(cls, v: str) -> str:
        parsed = urlparse(v)
        if parsed.scheme not in ('mqtt', 'mqtts'):
            raise ConfigValidationError(
                field='url',
                details="Протокол должен быть mqtt:// или mqtts://"
            )
        if parsed.netloc == '':
            raise ConfigValidationError(
                field='url',
                details="Отсутствует адрес брокера"
            )

        return v

    @model_validator(mode='after')
    def validate_ssl_requirements(self) -> Self:
        if not self.url:
            raise ConfigValidationError(
                field="url",
                details=f"Отсутствие поля 'url'."
            )
        parsed = urlparse(self.url)
        if parsed.scheme == 'mqtts':
            if not self.ssl:
                raise ConfigValidationError(
                    field='ssl',
                    details="SSL конфигурация обязательна для MQTTS"
                )

            for file_type in ('certFile', 'keyFile'):
                if not Path(getattr(self.ssl, file_type)).exists():
                    raise ConfigValidationError(
                        field=f'ssl.{file_type}',
                        details=f"Файл {getattr(self.ssl, file_type)} не найден"
                    )

        return self

    @classmethod
    def from_file(cls, config_file: str) -> Self:
        """Загрузка конфигурации из JSON-файла"""
        try:
            file = Path(config_file)
            if not file.exists():
                raise FileNotFoundError(f"Файл не найден: {config_file}")
            return cls.model_validate_json(file.read_text())

        except FileNotFoundError as e:
            raise ConfigValidationError(
                field='config_file',
                details=f"Файл конфигурации не найден: {config_file}"
            ) from e

        except json.JSONDecodeError as e:
            raise ConfigValidationError(
                field="config_file",
                details="Некорректный JSON формат"
            ) from e

        except ValidationError as e:
            raise ConfigValidationError(
                field="config_file",
                details=f"Ошибка валидации конфигурации: {e}"
            ) from e

class TagAttributes(BaseModel):
    # приходит от платформы
    prsActive: bool = Field(True)
    prsValueTypeCode: int = Field(..., ge=0, le=4)
    prsJsonConfigString: TagPrsJsonConfigStringFromPlatform

class PlatformConfig(BaseModel):
    prsActive: bool = True
    prsEntityTypeCode: int | None = None
    prsJsonConfigString: ConnectorPrsJsonConfigStringFromPlatform = ConnectorPrsJsonConfigStringFromPlatform()
    tags: dict[str, TagAttributes] = {}

    @field_validator('tags', mode='before')
    @classmethod
    def validate_tags_id(cls, v: dict[str, TagAttributes]) -> dict[str, TagAttributes]:

        for k in v.keys():
            try:
                uuid.UUID(str(k))
            except ValueError as e:
                raise ConfigValidationError(field='tags', details=f"id тега должен быть в виде GUID: {k}")

        return v

    @classmethod
    def _form_file_name(cls, connector_id: str) -> str:
        return f"platform_config_{connector_id}.json"

    @classmethod
    def from_file(cls, connector_id: str) -> Self:
        if (config_file := Path(cls._form_file_name(connector_id=connector_id))).exists():
            new_inst = cls.model_validate_json(config_file.read_text())
        else:
            new_inst = cls()
        new_inst.prsJsonConfigString.log.fileName = cls.default_log_file_name(connector_id)
        return new_inst

    @classmethod
    def default_log_file_name(cls, connector_id: str) -> str:
        return f"logs/prs_connector_{connector_id}.log"

    def save(self, connector_id: str) -> None:
        config_file = Path(self._form_file_name(connector_id=connector_id))
        data = self.model_dump_json(indent=2)
        config_file.write_text(data)


import sys
import asyncio
from time import time
# включаем в проект пакет для низкоуровневой работы с протоколом InduX
from py_indux import InduXClient, InduXException
# из пакета с базовым классом коннектора
# импортируем сам базовый класс и функцию,
# возвращающую текущую метку времени
from prs_connector_core import BaseConnector, ts, main

class InduXConnector(BaseConnector):

    # В общем случае, это единственный метод, который мы должны переопределить у базового класса.
    # Метод запускается базовым классом в виде отдельной задачи
    # каждый раз, когда от платформы приходит исправленная конфигурация
    # (не считая изменений, касающихся тегов, для них - отдельные сообщения).
    # Таким образом, метод _read_tags останавливается и после обновления конфигурации запускается вновь
    async def _read_tags(self):

        async def get_data():
            # Для упрощения предположим, что коннектор читает все теги из источника
            # с одной частотой.
            # В противном случае придётся разбивать теги по группам в зависимости от частоты
            # чтения и запускать для каждой группы свою задачу чтения данных из источника.
            interval = self._config_from_platfrom.prsJsonConfigString.source["frequency"]
            while True:

                # Зафиксируем время начала выполнения чтения.
                start_time = time()

                # Заготовка для пакета данных.
                data = {"data": []}
                # Время в платформе хранится как количество микросекунд, прошедших, начиная
                # с 01-01-1970 00:00:00 UTC.
                # Функция ts() текущую метку времени в указанном формате.
                # Предполагаем, что источник данных не возвращает метку времени вместе со значением,
                # поэтому мы должны присвоить её сами.
                current_time = ts()
                # пробегаем циклом по всем тегам в конфигурации
                for tag_id, tag_config in self._config_from_platfrom.tags:

                    # читаем значение каждого тега из конфигурации
                    value = await self._client.read(tag_config.prsJsonConfigString.source["name"])
                    # формируем пакет данных
                    data["data"].append(

                        {
                            "tagId": tag_id,
                            "data": [

                                [value, current_time]

                            ]

                        }

                    )

                # Помещаем сформированный пакет данных в очередь сообщений.
                self._data_queue.put_nowait(data)

                self._logger.info("Цикл чтения данных завершён.")

                # Считаем, сколько времени заняло чтение данных.
                elapsed = time() - start_time
                wait_time = max(0, interval - elapsed)
                # Делаем задержку между циклами чтения.
                await asyncio.sleep(wait_time)

        while True:
            try:
                # Соединяемся с источником данных
                source = self._config_from_platfrom.prsJsonConfigString.source
                self._client = InduXClient(

                    port=source["port"],
                    baudrate=source["baudrate"],
                    bytesize=source["bytesize"],
                    parity=source["parity"],
                    stopbits=source["stopbits"],
                    timeout=source["timeout"]

                )

                await self._client.connect()

                self._logger.info("Соединение с источником установлено.")

                # Запускаем цикл чтения данных.
                await get_data()
            except InduXException as ex:
                self._logger.exception(f"Ошибка чтения данных: {ex}")

                # При ошибке чтения записываем в каждый тег значение None
                # с кодом качества = 102, что обозначает разрыв связи коннектора
                # с источником данных.
                data = {"data": []}
                current_time = ts()
                for tag_id, tag_config in self._config_from_platfrom.tags:

                    data["data"].append(

                        {"tagId": tag_id, "data": [[None, current_time, 102]]}

                    )

                self._data_queue.put_nowait(data)

if __name__ == "__main__":
    main(InduXConnector)


import pytest
import json
from pathlib import Path
from prs_connector_core.config import ConnectorConfig
from prs_connector_core.exceptions import ConfigValidationError
import uuid

def test_connector_config_ssl(tmp_path, monkeypatch):

    config_file = tmp_path / "config.json"
    config_content = {
        "id": "550e8400-e29b-41d4-a716-446655440000",
        "url": "mqtts://localhost",
        "ssl": {"certFile": "cert.pem", "keyFile": "key.pem", "certsRequired": "CERTS_NONE"}
    }

    config_file.write_text(json.dumps(config_content))

    def mock_exists(path):
        # Разрешаем проверку для самого файла конфигурации
        if path == config_file:
            return True

        # Для сертификатов всегда возвращаем True
        return path.name in ("cert.pem", "key.pem")

    monkeypatch.setattr(Path, "exists", mock_exists)


    config = ConnectorConfig.from_file(str(config_file))
    assert config.ssl is not None
    assert config.ssl.certFile == "cert.pem"
    assert config.ssl.keyFile == "key.pem"

def test_connector_config_from_nonexistent_file(mocker):
    # Используем mock_open для эмуляции отсутствия файла
    mock_file = mocker.MagicMock()
    mock_file.exists.return_value = False
    mock_file.read_text.side_effect = FileNotFoundError("File not found")

    # Мокаем только конкретный путь
    mocker.patch("pathlib.Path", return_value=mock_file)

    with pytest.raises(ConfigValidationError):
        ConnectorConfig.from_file('nonexistent_file')

def test_connector_config_from_file(tmp_path, monkeypatch):

    config_file = tmp_path / "config.json"
    config_content = {"id": "550e8400-e29b-41d4-a716-446655440000", "url": "mqtt://localhost"}

    config_file.write_text(json.dumps(config_content))

    config = ConnectorConfig.from_file(str(config_file))


    assert str(config.id) == "550e8400-e29b-41d4-a716-446655440000"
    assert config.url == "mqtt://localhost"

def test_connector_config_id():
    id = "550e8400-e29b-41d4-a716-446655440000"
    config = ConnectorConfig(
        id=id,
        url="mqtt://localhost"
    )
    assert str(config.id) == id

    id = "550e8400-e29b-41d4-a716-44665544000"
    with pytest.raises(ConfigValidationError):
        ConnectorConfig(id=id, url="mqtt://localhost")

def test_connector_config_protocol():
    config = ConnectorConfig(
        id=str(uuid.uuid4()),
        url="mqtt://localhost"
    )
    assert config.url.startswith("mqtt://")

    with pytest.raises(ConfigValidationError):
        ConnectorConfig(id=str(uuid.uuid4()), url="http://localhost")

def test_connector_config_server():
    config = ConnectorConfig(
        id=str(uuid.uuid4()),
        url="mqtt://localhost"
    )
    assert config.url

    with pytest.raises(ConfigValidationError):
        ConnectorConfig(id=str(uuid.uuid4()), url="mqtt://")



from uuid import uuid4
from prs_connector_core.config import PlatformConfig, ConnectorPrsJsonConfigStringFromPlatform

def test_platform_config_from_file():
    connector_id = str(uuid4())
    plt_config = PlatformConfig.from_file(connector_id=connector_id)
    con_json_config_str = ConnectorPrsJsonConfigStringFromPlatform()
    con_json_config_str.log.fileName = f"logs/prs_connector_{connector_id}.log"
    assert plt_config.prsJsonConfigString == con_json_config_str
    assert plt_config.tags == {}